/* Design Implementation - Design Component - Model Branch
 *
 * Design.cpp
 * fwk4gps version 1.1
 * gam670/dps905
 * January 24 2011
 * copyright (c) 2011 Chris Szalwinski 
 * distributed under TPL - see ../Licenses.txt
 */

#include "iContext.h"        // for the Context Interface
#include "iText.h"           // for the Text Interface
#include "iSound.h"          // for the Sound Interface
#include "iLight.h"          // for the Light Interface
#include "iObject.h"         // for the Object Interface
#include "iTexture.h"        // for the Texture Interface
#include "iCamera.h"         // for the Camera Interface
#include "iScene.h"          // for the Scene Interface
#include "iHUD.h"            // for the HUD Interface
#include "iUtilities.h"      // for strcpy()
#include "iPhysics.h"
#include "Design.h"          // for the Design class declaration
#include "MathDefinitions.h" // for MODEL_Z_AXIS
#include "ModelSettings.h"   // for FLOOR, MOUSE_BUTTON_SCALE, ROLL_SPEED
#include "iSimpleCollisionSpace.h"
#include "Configuration.h"
#include "CollisionGeometry.h"
#include "COBB.h"
#include "CSphere.h"
#include <sstream>

//-------------------------------- Design -------------------------------------
//
// The Design Component implements the design within the model branch
//
// CreateDesign creates the Design Object
//
iDesign* CreateDesign(iContext* c, iScene* s, iHUD* h) {

	return new Design(c, s, h);
}

// constructor initializes the object and texture pointers and the reference
// time
//
Design::Design(iContext* c, iScene* s, iHUD* h) : context(c),
 scene(s), hud(h) {
	cs_=CreateSimpleCollisionSpace(60);
	CollisionGeometry::setGlobalCollisionSpace(cs_);


    // initialize reference time
	lastUpdate = 0;
}

//converts strings to wide character strings
wchar_t* MultiCharToUniChar(const char* mbString){
	int len = 0;
	len = (int)strlen(mbString) + 1;
	wchar_t *ucString = new wchar_t[len];
	mbstowcs(ucString, mbString, len);
	return ucString;
}

// initialize creates the objects and textures in the initial scene
// and initializes the reference time
//
void Design::initialize(int now) {

    // general parameters
    //
    context->set(GF_FR_NEAR, NEAR_CLIPPING);
    context->set(GF_FR_FAR, FAR_CLIPPING);
    context->set(GF_FR_FOV, FIELD_OF_VIEW);
    Colour white(1, 1, 1);
	topLight = CreateDistantLight(white, white, white, true);
	topLight->rotatex(3.14159f / 1.4f);
	context->set(GF_LT_BKGD, Vector(LIGHT_AMBIENT,LIGHT_AMBIENT,LIGHT_AMBIENT));

    //create the object textures
	iTexture* duckTex = CreateTexture(L"duckCM.tga");
    iTexture* katanaTex = CreateTexture(L"katana.bmp");
	iTexture* monitorTex = CreateTexture(L"monitor_Blue.bmp");
	iTexture* zombieTex = CreateTexture(L"zombie_texture.bmp");

	// camera at a distance - in lhs coordinates
    (camera = CreateCamera(context))->translate(80,50,0);
    camera->setRadius(17.8f);
	//camera->rotatex(3.14159f/4.0f);
	camera->rotatey(ToRadians(90));
	platform_=CreatePhysicsBox(-20,-20,-20,20,20,20,Colour(0.1f,0.1f,0.9f),8);
	cannon_ = CreatePhysicsCylinder(5, 40, 100,Colour(0.9f,0.1f,0.1f),1000); 
	//cannon_=CreatePhysicsBox(-5,0,-5,5,40,5,Colour(0.9f,0.1f,0.1f),1000);
	cannon_->attachTo(platform_,0,0);
	cannon_->translate(0,20,0);
	camera->attachTo(platform_->bv());//attach camera to the platform

	//create the bottle projectile and move it so that the player cannot see it (this mesh will be used later on this saves us having to reload it from the file every time)
	bottleCannon = CreatePhysicsMesh(5, 40,L"water_bottle.obj",Colour(0.9f,0.1f,0.1f),5,true);
	bottleCannon->translate(0,500,0);

	//create the duck projectile and move it so that the player cannot see it (this mesh will be used later on this saves us having to reload it from the file every time)
	cannonFodder = CreatePhysicsMesh(5, 40,L"duck_cannonball.obj",Colour(0.9f,0.1f,0.1f),7,true);
	cannonFodder->bv()->attach(duckTex);
	cannonFodder->translate(0,500,0);

	//create the ball projectile and move it so that the player cannot see it (this mesh will be used later on this saves us having to reload it from the file every time)
	ball =  CreatePhysicsMesh(5, 40,L"miniball.obj",Colour(0.9f,0.1f,0.1f),30,true);

	//create the katana projectile and move it so that the player cannot see it (this mesh will be used later on this saves us having to reload it from the file every time)
	katana = CreatePhysicsMesh(5, 40,L"katana.obj",Colour(0.9f,0.1f,0.1f),10,true);
	katana->bv()->attach(katanaTex);
	katana->translate(0,500,0);

	//by detault the cannonball is set to the ball mesh
	cannonball_ = ball;
 	cannonball_->translate(0,60,0);
	cannonball_->attachTo(cannon_,0,0);

	//first zombie
	//the zombie model was provided by pixel house studios and is under a free license
	zombie = CreatePhysicsMesh(30, 100,L"zombie_small.obj",Colour(0.9f,0.1f,0.1f),90,false);
	zombie->bv()->attach(zombieTex);
	zombie->translate(-50,30,0);
	zombie->rotatey(ToRadians(180));
	zombie->setCenterOfMass(Vector(0.0,25.0,0.0)); //need to adjust center of mass otherwise the collision mesh is not centered around the center of the zombie mesh
	zombie->setCollision(CreateCSphere(zombie,30));

	//second zombie

	iPhysics* zombie2 = zombie->clone();
	zombie2->bv()->attach(zombieTex);
	zombie2->translate(-200,50,160);
	zombie2->rotatey(ToRadians(180));
	zombie2->setCenterOfMass(Vector(0.0,25.0,0.0)); //need to adjust center of mass otherwise the collision mesh is not centered around the center of the zombie mesh
	zombie2->setCollision(CreateCSphere(zombie2,30));

	//first duck
	duck = CreatePhysicsMesh(30, 100,L"big_duck.obj",Colour(0.9f,0.1f,0.1f),6,false);
	duck->bv()->attach(duckTex);
	duck->translate(-100,30,0);
	duck->setCollision(CreateCSphere(duck,30));

	//second duck
	iPhysics* duck2 = duck->clone(); 
	duck2->bv()->attach(duckTex);
	duck2->translate(-180,60,40);
	duck2->setCollision(CreateCSphere(duck2,30));

	//first monitor
	monitor = CreatePhysicsMesh(20, 30,L"monitor.obj",Colour(0.9f,0.1f,0.1f),10,false);
	monitor->bv()->attach(monitorTex);
	monitor->translate(-90,80,80);
	monitor->setCollision(CreateCSphere(monitor,23));

	//Second monitor
	iPhysics* mon2 = monitor->clone();
	mon2->translate(-120,0,250);
	mon2->setCollision(CreateCSphere(mon2,23));

	//first lemonade stand
	lemon_stand = CreatePhysicsMesh(20, 30,L"lemonade_stand.obj",Colour(0.9f,0.1f,0.1f),10,false);
	lemon_stand->translate(-140,80,200);
	lemon_stand->setCenterOfMass(Vector(0.0,-30.0,0.0));
	lemon_stand->setCollision(CreateCSphere(lemon_stand,50));

	//second leomonade stand
	iPhysics* lemon2 = lemon_stand->clone();
	lemon2->translate(-120,80,270);
	lemon2->setCenterOfMass(Vector(0.0,-30.0,0.0));
	lemon2->setCollision(CreateCSphere(lemon2,50));
	
	//create ceiling
	ceiling=CreatePhysicsBox(-240,-10,-70,30,5,470,Colour(0.1f,0.9f,0.1f),10000,true);
	ceiling->translate(0,200,0);

	//create floor
	bigGreenBox_=CreatePhysicsBox(-500,-10,-100,0,5,500,Colour(0.1f,0.9f,0.1f),10000,true);
	bigGreenBox_->translate(60,-5,0);

	//create and position left wall
	orange_wall=CreatePhysicsBox(-100,10,-100,-90,110,500,Colour(0.9f,0.5f,0.0f),10000,true);
	orange_wall->rotatey(ToRadians(90));
	orange_wall->translate(0,0,-310);

	//create and position the floating tilted left wall
	tilted_air_wall1=CreatePhysicsBox(-100,10,-100,-90,110,500,Colour(0.9f,0.5f,0.0f),10000,true);
	tilted_air_wall1->rotatey(ToRadians(90));
	tilted_air_wall1->translate(0,107,-310);
	tilted_air_wall1->rotatex(ToRadians(45));

	//create and position the floating tilted right wall
	tilted_air_wall2=CreatePhysicsBox(-100,10,-100,-90,110,500,Colour(0.3f,0.5f,0.0f),10000,true);
	tilted_air_wall2->rotatey(ToRadians(90));
	tilted_air_wall2->translate(0,107,310);
	tilted_air_wall2->rotatex(ToRadians(-45));

	//create and position the floating tilted back wall
	tilted_air_wall3=CreatePhysicsBox(-100,10,-80,-90,110,450,Colour(0.1f,0.7f,0.9f),10000,true);
	tilted_air_wall3->translate(-200,105,0);
	tilted_air_wall3->rotatez(ToRadians(-45));

	//create and position the right wall
	purple_wall=CreatePhysicsBox(-100,10,-100,-90,110,500,Colour(0.5f,0.0f,0.5f),10000,true);
	purple_wall->rotatey(ToRadians(90));
	purple_wall->translate(0,0,310);

	//create and position the back wall
	bigBlueBox_=CreatePhysicsBox(-100,10,-100,-90,110,500,Colour(0.1f,0.1f,0.9f),10000,true);
	bigBlueBox_->translate(-200,0,0);

    CreateGrid(-150, FLOOR - 3, 150, 16,"opaque",
	Colour(0.1f, 0.1f, 0.1f,1))->translate(-60,0,0);
	lastfired_=now;
    lastUpdate = now;
	nf = NULL;
	fire = false;
	cannonPower = 0;
	totalScore = 0;
	ballSwitched = true;
	duckSwitched = false;
	katanaSwitched = false;
	bottleSwitched = false;
	firstUpdate = true;
	highestScore = 0;
	lastSwitched = 0;
	powerScore = CreateText(0.02f, 0.2f, 1, 0.99f, TXT_CENTER,L"power:0\n" L"score:0\n" L"High score:0\n" L"Use 1,2,3 and 4 to change the type of projectile\n" L"use the arrow keys to move the platform\n");
	hud->add(powerScore);


}

// reset changes the context if any change has been requested
// and reinitializes the reference time
//
void Design::reset(int now) {

    lastUpdate = now;
}

// restore re-initializes the reference time
//
void Design::restore(int now) {

    lastUpdate = now;
}

// update updates the position and orientation of each object for time "now"
// according to the keys pressed - the coordinate system for user actions is
// the left handed system with the x axis to the right, the y axis to the top 
// and the z axis into the screen
//
void Design::update(int now) {

    int delta;
    int dr = 0;  // roll the right box around its x axis
    int ds = 0;  // spin the right box around its y axis
    int dt = 0;  // roll the top   box around its z axis
    int dw = 0;  // roll the spot light around the x axis
	float vrn, J;
	Vector force,n,relativeVelocity;
	Vector g1deltap, g2deltap;
	const CollisionContact* cc;
   delta  = now - lastUpdate;

   //sets the current projectile to the ball
   if(context->pressed(SWITCH_CANNONBALL) && ballSwitched == false && (now-lastSwitched)>UNITS_PER_SEC/2.0f){
	   ball->bv()->setWorld(cannonball_->bv()->world());
	   cannonball_->attachTo(NULL,0,0);
	   cannonball_->translate(0,500,0);
	   cannonball_ = ball;
	   cannonball_->attachTo(cannon_,0,0);
	   ballSwitched = true;
	   duckSwitched = false;
	   katanaSwitched = false;
	   bottleSwitched = false;
	   lastSwitched = now;
   }
   //sets the current projectile to the duck
   if(context->pressed(SWITCH_DUCK) && duckSwitched == false && (now-lastSwitched)>UNITS_PER_SEC/2.0f){
	   cannonFodder->bv()->setWorld(cannonball_->bv()->world());
	   cannonball_->attachTo(NULL,0,0);
	   cannonball_->translate(0,500,0);
	   cannonball_ = cannonFodder;
	   cannonball_->attachTo(cannon_,0,0);
	   Vector test = cannonball_->velocity();	  
	   ballSwitched = false;
	   duckSwitched = true;
	   katanaSwitched = false;
	   bottleSwitched = false;
	   lastSwitched = now;
   }
   //sets the current projectile to the katana
   if(context->pressed(SWITCH_KATANA) && katanaSwitched == false && (now-lastSwitched)>UNITS_PER_SEC/2.0f){
	   katana->bv()->setWorld(cannonball_->bv()->world());
	   cannonball_->attachTo(NULL,0,0);
	   cannonball_->translate(0,500,0);
	   cannonball_ = katana;
	   cannonball_->attachTo(cannon_,0,0);
	   ballSwitched = false;
	   duckSwitched = false;
	   katanaSwitched = true;
	   bottleSwitched = false;
	   lastSwitched = now;
   }
   //sets the current projectile to the water bottle
   if(context->pressed(SWITCH_WATER) && bottleSwitched == false && (now-lastSwitched)>UNITS_PER_SEC/2.0f){
	   bottleCannon->bv()->setWorld(cannonball_->bv()->world());
	   cannonball_->attachTo(NULL,0,0);
	   cannonball_->translate(0,500,0);
	   cannonball_ = bottleCannon;
	   cannonball_->attachTo(cannon_,0,0);
	   ballSwitched = false;
	   duckSwitched = false;
	   katanaSwitched =false;
	   bottleSwitched = true;
	   lastSwitched = now;
   }
   if(context->pressed(FIRE_CANNONBALL)&& (now-lastfired_)>UNITS_PER_SEC/10.0f){
	    lastfired_=now;
		if(fire == false){
	    nf=cannonball_->clone();
		nf->attachTo(cannonball_,100,0);
		fire = true;
		}
		//increments the cannon power while the space bar is held down
		if(cannonPower <1000){
		cannonPower +=50;
		}
   }
   //checks if the fire button has been released and then fires the projectile
   else if(fire == true && (now-lastfired_)>UNITS_PER_SEC/10.0f){
	   flyinglist_.insertAtFront(nf);
	   if(cannonball_ == ball)
	      nf->setCollision(CreateCSphere(nf,2.5));//creates a sphere collision mesh roughly the size of the ball
	   else if(cannonball_ == cannonFodder)
		  nf->setCollision(CreateCSphere(nf,3.5));//creates a sphere collision mesh roughly the size of the duck
	   else if(cannonball_ == katana)
		  nf->setCollision(CreateCSphere(nf,7));//creates a sphere collision mesh roughly the size of the katana
	   else if(cannonball_ == bottleCannon)
	      nf->setCollision(CreateCSphere(nf,4.5));//creates a sphere collision mesh roughly the size of the bottle
	   fire = false;
	   nf->attachTo(NULL,100,0);
	   nf->addimpulseForce((cannonPower*4) *(nf->velocity()));//adds an impulse to the projectile that is modified by how much you have charged the cannon
	   cannonPower = 0;
   }
   if(context->pressed(CANNON_ROTATE_LEFT)){
	   cannon_->rotate(Vector(-1,0,0),(delta/1000.0f)*(3.14159f/4.0f));
   }
   if(context->pressed(CANNON_ROTATE_RIGHT)){
	   cannon_->rotate(Vector(1,0,0),(delta/1000.0f)*(3.14159f/4.0f));
   }
   //moves the cannon forward on the X axis
   if(context->pressed(CANNON_MOVE_FWD)){
	   platform_->translate(-1,0,0);
   }
   //moves the cannon backwords on the X axis
   if(context->pressed(CANNON_MOVE_BACK)){
	   platform_->translate(1,0,0);
   }
   //moves the cannon left on the Z axis
   if(context->pressed(CANNON_MOVE_LEFT)){
	   platform_->translate(0,0,-1);
   }
   //moves the cannon right on the Z axis
   if(context->pressed(CANNON_MOVE_RIGHT)){
	   platform_->translate(0,0,1);
   }
   if(context->pressed(CANNON_ROTATE_FWD)){
	   cannon_->rotate(Vector(0,0,1),(delta/1000.0f)*(3.14159f/4.0f));   
   }
   if(context->pressed(CANNON_ROTATE_BACK)){
	   
	   cannon_->rotate(Vector(0,0,-1),(delta/1000.0f)*(3.14159f/4.0f));
   }
   if(context->pressed(CLOCKWISE_MOMENT)){
	   platform_->addMoment(Vector(0,2000,0));
   }
   if(context->pressed(COUNTERCLOCKWISE_MOMENT)){
	   platform_->addMoment(Vector(0,-2000,0));
   }
   cs_->populateContactList(delta/UNITS_PER_SEC);
   int nc=cs_->getNumContacts();

   //used to clear the score as it happens to incriment while objects are being loaded as they collide while they are being positioned
   if(nc == 0  && firstUpdate){
	   firstUpdate = false;
	   totalScore = 0;
   }
   if(!firstUpdate){
   //if a certain amount of time has elapsed since the last collision set the high score to the total score and wipe the total score
   if(nc == 0 && (now - lastCollision) > UNITS_PER_SEC/2.0f){
	   if(totalScore > highestScore)
	      highestScore = totalScore;
	   totalScore = 0;
   }
   totalScore += (nc*100);
   std::stringstream ss;
   //update the HUD with the new scoring information
   ss << "power:" << cannonPower << "\n" << "score:" << totalScore << "\n" "High score:" << highestScore << "\n" << "Use 1,2,3 and 4 to change the type of projectile\n" << "use the arrow keys to move the platform\n";
   std::string s = ss.str();
   const char* toConvert = s.c_str();
   wchar_t* hud_text = MultiCharToUniChar(toConvert); 
   powerScore->set(hud_text);
   }
   
   if(nc!=0){
	   lastCollision = now;
    	   cc=cs_->getContactList();
	   if(cc->g1->getPhysics()!=bigBlueBox_ && cc->g1->getPhysics()!=bigGreenBox_){
    		  n=cc->normal;

		 Vector poa1 = cc->pos - cc->g1->getPhysics()->worldCentreOfMass();
		  Vector poa2 = cc->pos - cc->g2->getPhysics()->worldCentreOfMass();

		  Matrix piI1 = cc->g1->getPhysics()->rotation().transpose() * cc->g1->getPhysics()->getInvMomentOfIntertia() * cc->g1->getPhysics()->rotation();
		  Matrix piI2 = cc->g2->getPhysics()->rotation().transpose() * cc->g2->getPhysics()->getInvMomentOfIntertia() * cc->g2->getPhysics()->rotation();

		 relativeVelocity = cc->g1->getPhysics()->velocity() - cc->g2->getPhysics()->velocity();

	    // normal component of the relative velocity
		 vrn = dot(relativeVelocity, n);

	    // magnitude of the impulse at collision
	
		 J = - vrn * 2.0f / (1.0f / cc->g1->getPhysics()->mass() + 1.0f / cc->g2->getPhysics()->mass());
		 Vector impulse = J * n;
		
        cc->g1->getPhysics()->incrimentAngularVelocity(cross(poa1, impulse) * piI1);//my attempt at making the objects rotate when hit at a certain angle, it doesn't make them rotate but their movment does seem smoother
        cc->g2->getPhysics()->incrimentAngularVelocity( cross(poa2, -impulse) * piI2);//my attempt at making the objects rotate when hit at a certain angle, it doesn't make them rotate but their movment does seem smoother
	    // force generated by the impulse
	    force = J * n / (float(delta)/float(UNITS_PER_SEC));

		 // apply the force to both objects

		cc->g1->getPhysics()->addimpulseForce(force);
		cc->g2->getPhysics()->addimpulseForce(-1*force);

		//push objects apart so that it doesn't keep colliding
        float massTotal = (cc->g1->getPhysics()->mass() + cc->g2->getPhysics()->mass()) * 0.85f;
        float rbd1mf = cc->g2->getPhysics()->mass() / massTotal;
        float rbd2mf = cc->g1->getPhysics()->mass() / massTotal;
        Vector g1deltap (-1*(rbd1mf * cc->depth * n));
        Vector g2deltap (rbd2mf * cc->depth * n);
 		cc->g1->getPhysics()->translate(g1deltap.x,g1deltap.y,g1deltap.z);
 		cc->g2->getPhysics()->translate(g2deltap.x,g2deltap.y,g2deltap.z);
	   }
	   if(cc->g2->getPhysics()!=bigBlueBox_ && cc->g2->getPhysics()!=bigGreenBox_){
 		  n=cc->normal;

		  Vector poa1 = cc->pos - cc->g1->getPhysics()->worldCentreOfMass();
		  Vector poa2 = cc->pos - cc->g2->getPhysics()->worldCentreOfMass();

		  Matrix piI1 = cc->g1->getPhysics()->rotation().transpose() * cc->g1->getPhysics()->getInvMomentOfIntertia() * cc->g1->getPhysics()->rotation();
		  Matrix piI2 = cc->g2->getPhysics()->rotation().transpose() * cc->g2->getPhysics()->getInvMomentOfIntertia() * cc->g2->getPhysics()->rotation();
		 relativeVelocity = cc->g1->getPhysics()->velocity() - cc->g2->getPhysics()->velocity();

	    // normal component of the relative velocity
		 vrn = dot(relativeVelocity, n);

	    // magnitude of the impulse at collision
		 J = - vrn * 2.0f / (1.0f / cc->g1->getPhysics()->mass() + 1.0f / cc->g2->getPhysics()->mass());
		 Vector impulse = J * n;
	
        cc->g1->getPhysics()->incrimentAngularVelocity(cross(poa1, impulse) * piI1);//my attempt at making the objects rotate when hit at a certain angle, it doesn't make them rotate but their movment does seem smoother
        cc->g2->getPhysics()->incrimentAngularVelocity( cross(poa2, -impulse) * piI2);//my attempt at making the objects rotate when hit at a certain angle, it doesn't make them rotate but their movment does seem smoother

		   // force generated by the impulse
	    force = J * n / (float(delta)/float(UNITS_PER_SEC));

		 // apply the force to both objects

		cc->g1->getPhysics()->addimpulseForce(force);
		cc->g2->getPhysics()->addimpulseForce(-1*force);

		//push objects apart
        float massTotal = (cc->g1->getPhysics()->mass() + cc->g2->getPhysics()->mass()) * 0.85f;
        float rbd1mf = cc->g2->getPhysics()->mass() / massTotal;
        float rbd2mf = cc->g1->getPhysics()->mass() / massTotal;
		g1deltap=Vector (-1*(rbd1mf * cc->depth * n));
        g2deltap=Vector (rbd2mf * cc->depth * n);
 		cc->g1->getPhysics()->translate(g1deltap.x,g1deltap.y,g1deltap.z);
 		cc->g2->getPhysics()->translate(g2deltap.x,g2deltap.y,g2deltap.z);
	   }
   }
   cannonball_->update(delta);
   cannon_->update(delta);
   platform_->update(delta);
   flyinglist_.goStart();
   while(flyinglist_.curr()){
	   iPhysics* cb=flyinglist_.curr()->data();
 	   cb->update(delta);
	   if(cb->position().y < FLOOR-3){
		   cs_->remove(cb->collisionGeometry());
		   flyinglist_.removeCurrent();
		   delete cb;
	   }
	   else{
		   flyinglist_.goNext();
	   }
   }
    lastUpdate = now;
}

// draw draws the different objects under their own rules
//
void Design::draw() {

    scene->draw(MIRROR);
    scene->draw(SKYBOX);
    scene->draw(OPAQUE_OBJECT);
    scene->draw(TRANSLUCENT_OBJECT);
    hud->draw(HUD_ALPHA);
}

void Design::preDraw() {

    scene->preDraw(MIRROR);
}
